#pragma once

#include "OptimizerBase.h"
#include <iostream>

namespace mdd {
	class OptimizerEvolutionary : public OptimizerBase {
	protected:
		static int random_num(double min, double max, double inc = 1.0)
		{
			int range = (max - min) / inc + 1;
			return min + inc * (std::rand() % range);
		}

		Permutation generatePermutation();
		Permutation combine(Permutation par1, Permutation par2);

		std::vector<Permutation> _children;
		std::vector<Permutation> _bests;
		int _min_generations;
		double _max_fitness;
		size_t _grow_generation;
		size_t _converges;
		double _precision;

		void evolve(std::vector<Permutation> parents);
		void evaluate(size_t ignore_len = 0);
		std::vector<double> mutateGene(std::shared_ptr<IInput> input, std::vector<double> seed = std::vector<double>());
		bool configureChild(const json& config) override;

	public:
		OptimizerEvolutionary();
		double update() override;
		std::vector<Permutation> getBests();
		double evaluateFitness(const Permutation& ind);
	};
}