#include "OptimizerEvolutionary.h"

namespace mdd {
	void OptimizerEvolutionary::evaluate(size_t ignore_len)
	{
		//calculate fitness
		for (auto it = _children.begin() + ignore_len; it != _children.end(); ++it)
		{
			for (size_t j = 0; j < _inputs.size(); j++)
			{
				//std::cout << "set: " << j << ": " << it->dna[j] << std::endl;
				_inputs[j]->setValue() = it->dna[j];
			}
			auto opt = updateOutputs();
			callback_evaluate_opt(opt, *it);
			if (opt.module_state == state::STATE_ERROR) {
				_children.erase(it);
			}
			else {
				it->fitness = opt.opt_value;
			}
		}
		//find fitest
		double min = _children[0].fitness;
		for (size_t i = 1; i < _children.size(); i++)
		{
			if (round(min/_precision)*_precision > round(_children[i].fitness / _precision) * _precision) {
				min = _children[i].fitness;
				_bests.clear();
				_bests.push_back(_children[i]);
			}
			else if(round(min / _precision) * _precision == round(_children[i].fitness / _precision) * _precision)
			{
				bool found = false;
				for (size_t j = 0; j < _bests.size(); j++)
				{
					if (_children[i] == _bests[j])
					{
						found = true;
						break;
					}
				}
				if (!found)
				{
					_bests.push_back(_children[i]);
				}
			}
		}
		
		// json ret;
		// ret["step"] = step;
		// for (auto& ind : _bests)
		// {
		// 	json jind;
		// 	jind["fitness"] = ind.fitness;
		// 	jind["dna"] = ind.dna;
		// 	ret["bests"].push_back(jind);
		// }
		// json jcall;
		// jcall["operation"] = "add";
		// jcall["args"]["subject"] = getIdentifier();
		// jcall["args"]["object"] = ret;
		// _callback(jcall);
	}

	OptimizerEvolutionary::OptimizerEvolutionary()
		:OptimizerBase(R"JSON(
        {
            "converges":
			{
				"value": 0
			},
            "grow generation":
			{
            "value": 20
			},
            "max fitness":
			{
				"value": 0.0
			},
            "min generations": 
			{
				"value": -1
			},
            "precission":
			{
				"value": 0.001
			}
        })JSON")
	{
		key = "OptimizerEvolutionary";
		_converges = 0;
		_grow_generation = 20;
		_max_fitness = 0.0;
		_min_generations = -1;
		_precision = 0.001;
	}

	bool OptimizerEvolutionary::configureChild(const json& config) {
		bool found = false;
		auto jit = config.find("converges");
		if (jit != config.end())
		{
			json jval = jit.value()["value"];
			if (jval.is_string())
			{
				jval = json::parse(jval.get<std::string>());
			}
			_converges = jval.get<int>();
			_base_config["configure"]["converges"]["value"] = _converges;
			found = true;
		}
		jit = config.find("grow generation");
		if (jit != config.end())
		{
			json jval = jit.value()["value"];
			if (jval.is_string())
			{
				jval = json::parse(jval.get<std::string>());
			}
			_grow_generation = jval.get<int>();
			_base_config["configure"]["grow generation"]["value"] = _grow_generation;
			found = true;
		}
		jit = config.find("max fitness");
		if (jit != config.end())
		{
			json jval = jit.value()["value"];
			if (jval.is_string())
			{
				jval = json::parse(jval.get<std::string>());
			}
			_max_fitness = jval.get<int>();
			_base_config["configure"]["max fitness"]["value"] = _max_fitness;
			found = true;
		}
		jit = config.find("min generations");
		if (jit != config.end())
		{
			json jval = jit.value()["value"];
			if (jval.is_string())
			{
				jval = json::parse(jval.get<std::string>());
			}
			_min_generations = jval.get<int>();
			_base_config["configure"]["min generations"]["value"] = _min_generations;
			found = true;
		}
		jit = config.find("precision");
		if (jit != config.end())
		{
			json jval = jit.value()["value"];
			if (jval.is_string())
			{
				jval = json::parse(jval.get<std::string>());
			}
			_precision = jval.get<double>();
			_base_config["configure"]["precision"]["value"] = _precision;
			found = true;
		}
		return found;
	}

	double OptimizerEvolutionary::update()
	{
		step = -1;
		bool check;
		Permutation old_best;
		size_t same_counter = 0;
		if (_inputs.size() == 0)
		{
			std::cout << "ERROR: No optimizable inputs detected!" << std::endl;
			return -1.0;
		}
		_children.clear();
		_bests.clear();
		do
		{
			++step;
			std::cout << _children.size() << " | " << step << std::endl;
			if (_children.empty())
			{
				for (size_t i = 0; i < _grow_generation*2; i++)
				{
					_children.push_back(generatePermutation());
				}
				callback_add_step(_children);
				evaluate();
			}
			else {
				evolve(_children);
			}
			check = step < _min_generations || _bests[0].fitness > _max_fitness;
			if (!check && _converges > 0)
			{
				bool found = false;
				for (size_t i = 0; i < _bests.size(); i++)
				{
					if (_bests[i] == old_best)
					{
						found = true;
						break;
					}
				}
				if (found)
				{
					++same_counter;
				}
				else
				{
					old_best = _bests[0];
					same_counter = 0;
				}
				if (same_counter < _converges)
				{
					check = true;
				}
			}
		} while (check);
		
		return _bests[0].fitness;
	}

	std::vector<double> OptimizerEvolutionary::mutateGene(std::shared_ptr<IInput> input, std::vector<double> seed)
	{
	
		const limits limit = *(input->getLimits());
		auto ret = seed;
		
		if (!limit.elements.empty())
		{
			int i = (int)random_num(0, limit.elements.size() - 1);
			ret = limit.elements[i];
			return ret;
		}
		if (!limit.min.empty() && !limit.max.empty()) {

			bool check = true;
			do {
				if (!ret.empty())
				{
					int i = (int)random_num(0, limit.min.size() - 1);
					if (!limit.step.empty()) {
						//randomly generate step dx in [min, max]
						/*std::cout << limit["min"][i].get<double>() << " | ";
						std::cout << limit["max"][i].dump() << " | ";
						std::cout << limit["step"][i].dump() << " | ";*/
						ret[i] = random_num(limit.min[i], limit.max[i], limit.step[i]);
					}
					else {
						int m = (int)random_num(0, 1);
						if (m == 0)
						{
							ret[i] = limit.min[i];
						}
						else {
							ret[i] = limit.max[i];
						}
					}
				}
				else
				{
					for (size_t i = 0; i < limit.min.size(); i++)
					{
						if (!limit.step.empty()) {
							//randomly generate step dx in [min, max]
							ret.push_back(random_num(limit.min[i], limit.max[i], limit.step[i]));
						}
						else {
							int m = (int)random_num(0, 1);
							if (m == 0)
							{
								ret.push_back(limit.min[i]);
							}
							else {
								ret.push_back(limit.max[i]);
							}
						}
					}
				}
				
				if (!limit.rule.empty()) {
					//use rule to check
					exprtk::symbol_table<double> symbol_table;
					
					symbol_table.add_vector("val", ret);
					symbol_table.add_constants();
					exprtk::expression<double> func_expr;
					func_expr.register_symbol_table(symbol_table);

					exprtk::parser<double> parser;
					parser.compile(limit.rule, func_expr);
					double func_val = func_expr.value();
					if (func_val < 0.5)
					{
						check = false;
					}
					else {
						check = true;
					}
				}
			} while (!check);
			return ret; 
			
		}
		std::cout << "ERROR in GENE elements! (END)" << std::endl;
		return input->getValue();//ERROR
	}

	void OptimizerEvolutionary::evolve(std::vector<Permutation> parents)
	{
		////fittest should breed more
		//detect lower border
		double max = parents[0].fitness;
		for (size_t i = 1; i < parents.size(); i++)
		{
			if (max < parents[i].fitness) {
				max = parents[i].fitness;
			}
		}
		//detect lower border
		double sum = 0;
		for (size_t i = 0; i < parents.size(); i++)
		{
			sum += max - parents[i].fitness;
		}
		if (sum == 0)
		{
			sum = 1.0;
		}
		//multiply fitter parents
		std::vector<Permutation> gen_pool = parents;
		for (size_t i = 0; i < parents.size(); i++)
		{
			int additional = std::round((max - parents[i].fitness) / sum * 20);
			//std::cout << additional << std::endl;
			for (size_t j = 1; j < additional; j++)
			{
				gen_pool.push_back(parents[i]);
			}
		}
		//fittest parents will also be new childrens
		_children = _bests;
		
		//breed
		size_t init_len = _children.size();
		for (size_t i = 0; i < _grow_generation; i++)
		{
			int p1 = (int)random_num(0, gen_pool.size() - 1);
			int p2 = (int)random_num(0, gen_pool.size() - 1);
			_children.push_back(combine(gen_pool[p1], gen_pool[p2]));
		}
		callback_add_step(_children);
		evaluate(init_len);
	}

	OptimizerEvolutionary::Permutation OptimizerEvolutionary::combine(Permutation par1, Permutation par2)
	{
		size_t len = par1.dna.size();
		Permutation child;
		for (size_t i = 0; i < len; i++)
		{
			double p = random_num(0.0, 100.0, 1.0);
			if (p<45.0)
			{
				child.dna.push_back(par1.dna[i]);
			}
			else if(p<90.0){
				child.dna.push_back(par2.dna[i]);
			}
			else
			{
				child.dna.push_back(mutateGene(_inputs[i]));
			}
		}
		return child;
	}

	OptimizerEvolutionary::Permutation OptimizerEvolutionary::generatePermutation()
	{
		Permutation ret;
		for (size_t i = 0; i < _inputs.size(); i++)
		{
			ret.dna.push_back(mutateGene(_inputs[i]));
		}
		return ret;
	}

	std::vector<OptimizerEvolutionary::Permutation> OptimizerEvolutionary::getBests() {
		return _bests;
	}

	double OptimizerEvolutionary::evaluateFitness(const Permutation& ind) {
		for (size_t j = 0; j < _inputs.size(); j++)
		{
			_inputs[j]->setValue() = ind.dna[j];
		}
		auto opt = updateOutputs();
		return opt.opt_value;
	}
}
